# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import urllib
from string import Template

from pyramid.httpexceptions import HTTPFound, HTTPForbidden
from pyramid.response import Response
from pyramid.view import (
    view_config,
    view_defaults
    )

from .security import CanvasContext, ViewCanvas

log = logging.getLogger(__name__)


js_redirect_tpl = """<html>
  <body>
    <script>
      window.top.location = "%(location)s";
    </script>
  </body>
</html>"""


class OauthAccept(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request


class OauthDeny(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request


@view_defaults(route_name='facebook_canvas')
class FacebookCanvas(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @view_config(context=HTTPForbidden)
    def prompt_authorize(self):
        """Prompt user to authorize app:

        * User is not logged in
        * User is on apps.facebook.com/{namespace}
        """
        log.debug('prompt authorize')
        settings = self.request.registry.settings
        redirect_uri = urllib.quote_plus(
            "%s://apps.facebook.com%s" % (
                self.request.scheme,
                self.request.route_path('facebook_canvas_oauth')
                )
            )
        url = "%s/dialog/oauth/?client_id=%s&redirect_uri=%s&scope=%s" % (
            "https://www.facebook.com",
            settings['facebook.app_id'],
            redirect_uri,
            settings['facebook.scope'].replace(' ', '')
            )
        return Response(js_redirect_tpl % {'location': url})

    @view_config(permission=ViewCanvas)
    def canvas(request):
        """
        * User is logged in
        """
        raise NotImplementedError()


@view_defaults(route_name='facebook_canvas_oauth')
class FacebookOAuth(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _redirect_to_canvas(self):
        p = {
            'location': "%s://apps.facebook.com%s" % (
                self.request.scheme,
                self.request.route_path('facebook_canvas')
                )
            }
        return Response(js_redirect_tpl % p)

    @view_config(permission=ViewCanvas)
    def oauth_accept(self):
        log.debug('user accepts authorization')
        self.request.registry.notify(OauthAccept(self.context, self.request))
        return self._redirect_to_canvas()

    @view_config(context=HTTPForbidden, request_param="error")
    def oauth_deny(self):
        log.debug('user denies authorization')
        self.request.registry.notify(OauthDeny(self.context, self.request))
        return self._redirect_to_canvas()



# -*- coding: utf-8 -*-
import logging

from pyramid.authentication import CallbackAuthenticationPolicy
from pyramid.security import (
    IAuthenticationPolicy,
    Authenticated,
    Everyone,
    Allow
    )
from zope.interface import implements

from .lib import (
    decrypt_signed_request,
    FacebookSignatureException
    )

__all__ = ['FacebookAuthenticationPolicy', 'CanvasContext', 'ViewCanvas']

log = logging.getLogger(__name__)

ViewCanvas = 'view_canvas'


class CanvasContext(object):
    """Security context for facebook canvas routes.
    """
    def __init__(self, request):
        self.__acl__ = [
            (Allow, Authenticated, ViewCanvas),
        ]
        self._fb_data = None

    @property
    def facebook_data(self):
        """Contains facebook data provided in ``signed_request`` parameter.
        """
        return self._fb_data

    @facebook_data.setter
    def facebook_data(self, value):
        if self._fb_data is None:
            self._fb_data = value
        else:
            raise ValueError('Property can be set only once')


class FacebookAuthenticationPolicy(CallbackAuthenticationPolicy):
    implements(IAuthenticationPolicy)

    def __init__(self, secret_key, callback=None):
        log.debug('authentication policy init')
        self.secret_key = secret_key
        self.callback = callback

    def unauthenticated_userid(self, request):
        """
        """
        if request.params.get('signed_request') is None:
            return None

        context = request.context
        if context.facebook_data is None:
            # decrypt signed request once only
            try:
                context.facebook_data = decrypt_signed_request(
                    self.secret_key,
                    request.params["signed_request"]
                    )
            except FacebookSignatureException as e:
                context.facebook_data = dict()
                log.warn('Invalid signature %r', e)
                return None
        try:
            return long(context.facebook_data["user_id"])
        except KeyError as e:
            # user_id not in facebook_data => user has not authorized app
            return None
        except (TypeError, ValueError) as e:
            log.warn('Invalid user id %r', context.facebook_data["user_id"])
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import logging

from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator

from .security import FacebookAuthenticationPolicy

__all__ = ['includeme', 'main']

log = logging.getLogger(__name__)


def main(global_config, **settings):
    log.debug('global_config: %r\nsettings: %r', global_config, settings)
    config = Configurator(settings=settings)
    config.include(includeme)
    return config.make_wsgi_app()


def includeme(config):
    log.debug('config: %r', config)
    settings = config.registry.settings

    config.set_authentication_policy(
        FacebookAuthenticationPolicy(settings['facebook.secret_key'])
        )
    config.set_authorization_policy(ACLAuthorizationPolicy())

    facebook_path = '/%s' % settings['facebook.namespace']

    config.scan()

    config.add_route(
        'facebook_canvas',
        '%s/' % facebook_path,
        request_method='POST',
        request_param='signed_request',
        factory='pyramid_facebook.security.CanvasContext',
        )

    config.add_route(
        'facebook_canvas_oauth',
        '%s/oauth' % facebook_path,
        request_method='POST',
        request_param='signed_request',
        factory='pyramid_facebook.security.CanvasContext',
        )

    config.commit()



# -*- coding: utf-8 -*-
import base64
import hashlib
import hmac
import json
import logging

log = logging.getLogger(__name__)

_signed_request_algorithm = 'HMAC-SHA256'
_dict_translate = dict(zip(map(ord, u'-_'), u'+/'))


class FacebookSignatureException(Exception):
    """Raised when signed_request is invalid.
    """
    pass


def _base64_url_encode(inp):
    return unicode(base64.b64encode(inp, '-_').strip('=').encode('utf8'))


def _base64_url_decode(inp):
    """ Facebook base64 decoder.

    http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas """
    padding_factor = (4 - len(inp) % 4) % 4
    inp += '=' * padding_factor
    return base64.b64decode(inp.translate(_dict_translate))


def encrypt_signed_request(secret_key, data):
    """Encrypts data the way facebook does for permit testing. Adds algorithm
    key to dict.

    :param data: a dictionary of data to sign.
    """
    data = data.copy()
    data.update(algorithm=_signed_request_algorithm)

    payload = _base64_url_encode(json.dumps(data)).encode('utf8')
    signature = _base64_url_encode(
        hmac.new(
            secret_key.encode('utf8'),
            msg=payload.encode('utf8'),
            digestmod=hashlib.sha256
            ).digest()
        )
    return '%s.%s' % (signature, payload)


def decrypt_signed_request(secret_key, signed_request):
    """ Decrypts signed request sent by facebook. See
    http://developers.facebook.com/docs/authentication/canvas for more
    information.

    returns a dictionary with facebook parameters. """

    if signed_request.count('.') != 1:
        msg = 'signed_request badly formatted.'
        raise FacebookSignatureException(msg)

    signature, payload = signed_request.split('.')

    if not len(signature) > 0 or not len(payload) > 0:
        msg = 'Facebook signature and/or payload with null length.'
        raise FacebookSignatureException(msg)

    try:
        signature = _base64_url_decode(signature)
        data = json.loads(_base64_url_decode(payload))
    except:
        msg = 'Facebook signature or data base64 decoding failed.'
        raise FacebookSignatureException(msg)

    if not hasattr(data, 'get'):
        msg = 'Facebook signed request with invalid data.'
        raise FacebookSignatureException(msg)

    algorithm = data.get('algorithm')
    if not algorithm:
        msg = 'Facebook Signed Request without algorithm information.'
        raise FacebookSignatureException(msg)

    if algorithm.upper() != _signed_request_algorithm:
        msg = 'Facebook signature not encoded with HMAC-SHA256.'
        raise FacebookSignatureException(msg)

    expected_signature = hmac.new(
        secret_key.encode('utf8'),
        msg=payload.encode('utf8'),
        digestmod=hashlib.sha256
        ).digest()

    if signature != expected_signature:
        msg = 'Facebook signed request with invalid signature.'
        raise FacebookSignatureException(msg)

    return data
